use strict;
package DC_Test;
use Getopt::Long;
use utf8;    
binmode STDOUT, ":utf8";  
binmode STDIN, ":utf8";

require ("dc_check/dc_subs1.pl");
require ("dc_check/dc_version.pl");
#################### dc_checkvar1.pl ###############xxx
#
# check for undefined transducers
# check for unused and redefined transducers
# check for variables starting with $ or # also end with $ or #
# check if <> variables of an alphabet are set up properly
# 
# Used arrays:
#------------
# %repl contains fullest variable names
# @darr contains all inflection variables
# %barr contains all alphabet <...> variables with :
# @b1arr contains all alphabet #...# variables
# %b2arr contains all alphabet <...> variables w/o :
# %b3arr: #..# var and expanded part of it
# %carr contains all variable settings
# all saved names without <> or ##
#

my ($result, $input, $path, $version, $help, $did_check);
our $donotchkvar;

$result = GetOptions (
    "input=s" => \$input, # string
    "path=s" => \$path, # string
    "version" => \$version, # string
    "donotchkvar=i" => \$donotchkvar, # string
    "help" => \$help); # flag

if ($version){
   our $dc_version;
   print "dc_checkvar1.pl version=$dc_version\n";
   exit();
}

#print "dcv:$donotchkvar\n";

our (@arr, $as, $folytatas, %barr, %b2arr, %b3arr, @b1arr, %carr, $was_alphabet, %repl, @darr);
# pattern for transducer and #..# variables name (unused)
my $pat1 = "A-Za-z0-9_\$\#\/+&\(\)~\?\=";

my ($dirname, $slash, $fname) = get_fname($input, $path);
#my ($dirname, $slash, $fname) = get_fname($ARGV[0], $ARGV[1]);

open FILE, "<$fname" or die 'Can\'t open '.$fname;
binmode FILE, ":utf8";
our @input = <FILE>;
close(FILE);
#
#  check for unused and redefined transducers
#
check_transducers1();
#
# check for redefined transducers
#    
check_transducers2();
#
#  check for variables starting with $ or # also end with $ or #
#
check_variables();
#
#  replace all #name# variables with the fullest version
#    result in hash %repl
#
replace_hvariable_names();
#
#  collect all inflection variables in array @darr
#
collect_inflection_vars_in_darr();
#
# check if <> variables of an alphabet are set up properly
# %barr: in alphabet defined vars
# %carr: transducers
# @b1arr:#..# vars
# %b3arr: #..# var and expanded part of it
#
check_if_variable_setup_all_right();
#
# check_hash_variables()
# check if #..# = #..#<..><..> given
#    used <..> and #..# also  appear on an other place- otherwise likely typing error
check_hash_variables();
# at least one check
if(!$did_check){
  check_previous_alphabet();
}

sub check_if_variable_setup_all_right(){
#use utf8;
#binmode STDOUT, ":utf8";
my $debug = 0;
my $linenr;
foreach(@input){
   chomp;
   if($_ =~ /ALPHABET\s*=/){
     if($was_alphabet){
       check_previous_alphabet();
       %barr = ();
       %b2arr = ();
       %b3arr = ();
       %carr = ();
       @b1arr = ();
     }
     $was_alphabet = 1;
     my $seg = $_;
     #
     # put [#..#] variables and their expansion, typically
     #  <..> variables into %b3arr
     #  expanded part will be copied from %repl hash
     #
     if($_ =~ /\[/){
        @arr = $_ =~ /\[(.*?)\]/g;
        foreach(@arr){
          my $levagott = $_;
          $levagott =~ s/^#//; #remove leading hash
          $levagott =~ s/#$//; #remove trailing hash
          if($debug){print "found $_ levagott:'$levagott'\n";}
          if(defined($repl{$_})){
              my $seg1 =  $repl{$_};
              if($debug){print "$_ $seg1\n";}
              $b3arr{$_} = $seg1;
          } 
          elsif(defined($repl{$levagott})){
              my $seg1 =  $repl{$levagott};
              if($debug){print "$_ $seg1\n";}
              $b3arr{$levagott} = $seg1;
          } else{
                  if($debug> 2){ 
                     print "not found $_ in repl\n";
                     print_hash(\%repl, 'repl', 1);
                  }
           }
          
        }
     }
     @arr = split(/\s+/, $seg);
     $as = @arr;
     #
     # <blabla>:a and <blabla>:[abc] type correspondences
     #
     for(my $i = 0; $i < $as; $i++){
        if($arr[$i] =~ /<(.*)>:(.*)/){
            my ($mas, $prev);
           if(substr($2,0,1) eq '[' and substr($2, length($2)-1,1) eq ']'){
              my $seg = substr($2,1);
              $seg = substr($seg,0,length($seg)-1);
              my @marr = split(//, $seg);
              my $ms = @marr;
              for(my $i = 0; $i < $ms; $i++){
                if($marr[$i] eq '<' or $marr[$i] eq '#'){
                   my $end = '>';
                   if($marr[$i] eq '#') {$end = '#';}
                   #print "itt1 mas:'$mas'\n";
                   $mas .= $marr[$i];
                   $i++;
                   while($marr[$i] ne $end and $i < $ms){
                     $mas .= $marr[$i++];
                    # print "itt2 mas:'$mas'\n";
                   }
                   if($marr[$i] eq $end){$mas .= "$end ";}
                   else {$mas .= ' ';}
                } else{
                   $mas .= $marr[$i];
                 }
              } # for i
              #$mas = substr($mas, 0, length($mas)-1);
           } else {$mas =  $2;}
           my @ssarr = split(//, $mas);
           foreach(@ssarr){
             my $elem = $_;
             if($elem eq '<'){$elem = '<>';}
             elsif($elem eq '>') {next;}
             if(!defined($barr{$1})){
                $barr{$1} = $elem;
                 if($debug>2){print "1saved $mas\n";}
             }
             else{
                check_for_double_element($barr{$1}, $2, $1);
                 $barr{$1} .= ' '.$elem;
                 if($debug>2){print "2saved ' '.$mas\n";}
             }
           }
         } 
        #
        # individual <..> variables in alphabet
        #
        elsif ($arr[$i] =~ /<(.*?)>/){
            my @sarr = $arr[$i] =~ /<(.*?)>/g;
            foreach(@sarr){
              if(!defined($b2arr{$1})){
                $b2arr{$_} = 1;
                if($debug){print "3saved $_\n";}
              } else{
                ++$b2arr{$_};
              }
           }
        }
        #
        # individual #..# variables in alphabet
        #
        elsif ($arr[$i] =~ /#(.*)#/){
            my $seg1 = $1;
            my $found;
            foreach(@b1arr){
               if($_ eq $seg1){$found = 1; last;}
            }
            if(!$found){
               push(@b1arr, $seg1);
            }
        }     
     }
   } # ALPHABET
   #
   #  not an alphabet line...
   #
   else{
    # assignment using <=>
    if($_ =~ /<([^ ]*)>\s*<=>\s*([^ ]*)/){
           if($debug>2){
                print "1:$1 2:$2\n";
           }
           if(!defined($carr{$1})){
                $carr{$1} = $2;
           }
           else{
                $carr{$1} .= ' '.$2;
           }
     }
    # assignment using ^-> 
    elsif($_ =~ /\s*\^\->\s*.*?<(.*?)>/){
           if (!defined($b2arr{$1}) and !defined($barr{$1}) and !is_in_hash(\%b3arr, $1)){
               if($debug){
                  if($debug > 2) {
                      print_hash(\%barr, 'barr', 1);
                      print_hash(\%b2arr, 'b2arr', 1);
                  }
                  print_hash(\%b3arr, 'b3arr', 1);
                  #print_hash(\%repl, 'repl', 1);
               }
               print_message( "Warning: unknown <...> variable: $1 used near line $linenr\n");
          }
     } # ^-> found
   } # else
   ++ $linenr;
 } # foreach
}
#
#  Check if previous alphabet's definitions are correct
#
sub check_previous_alphabet(){
my ($name, $elements, $wassign, $size, $debug);
$debug = 0;
$did_check = 1;
if($debug){
  print_hash(\%barr,'barr',0);
  print_hash(\%carr,'carr',0);
} #debug
 while (($name, $elements) = each(%barr))
 {
    my (@ba, @ca, $bas, $cas);
    @ba = split(/\s+/, $elements);
    $bas = @ba;
    $cas = 0;
    if(defined($carr{$name})){
      @ca = split(/\s+/, $carr{$name});
      $cas = @ca;
    } else{
      if($bas > 1){
        print_message( "Definition missing for $name, element: $elements bas:$bas\n");
        $wassign = 1;
      }
    }
    if(!$wassign){
      if($cas > $bas-1){
         print_message( "Warning: Too many definitions for $name element:$elements defs:$carr{$name} cas:$cas bas:$bas \n");
      }elsif($cas < $bas -1){
         print_message( "Too few definitions for $name element:$elements defs:$carr{$name} cas:$cas bas:$bas \n");
      }
      if($cas){
         my $found = 0;
         foreach(@ca){
           if($bas){
              my $seg = $_;
              foreach(@ba){
               if($_ eq $seg){$found = 1; last;}
              }
              if(!$found){
                print_message( "Wrong definition for $name not found:$seg \n");
              }
           } #$bas
         } #@ca
      } #cas
      if($bas > 1){
         if(!$cas){ print_message( "Undefined $name element:$elements \n");}
      }
    } # wassign
    $wassign = 0;
 }
 check_if_alphabet_uses_all_inflect_variables();
}
#
# check if alphabet uses all inflect variables
# %repl contains fullest variable names
# @darr contains all inflection variables
# %barr contains all alphabet <...> variables
# @b1arr contains all alphabet #...# variables
# %b3arr: #..# var and expanded part of it
# %carr contains all variable settings
# all saved names without <> or ##
#
sub check_if_alphabet_uses_all_inflect_variables(){
   my ($debug, $name, $elements);
   if($debug > 1){while (($name, $elements) = each(%barr))
    {
   	  print "barr: $name has $elements\n";
    }
    foreach(@darr){
      print "darr:$_\n";
    }
   }# debug

   my @missing;
   foreach(@darr){
     my $seg1 = $_;
     my $found;
     # search in %barr first
    my ($name, $elements);
    while (($name, $elements) = each(%barr))
    {
	  if($debug){print "search for $seg1 in barr, $name has $elements\n";}
      if($name eq $seg1){
           $found = 1;
           if($debug){print "found $seg1\n";}
           #last;
      }
    }
    if($debug){if(!$found){ print "not found $seg1 yet\n";}}
    if(!$found){
      foreach(@b1arr){  
        if(defined($repl{$_})){
         my $seg2 = $repl{$_};
         if(good($seg1) and good($seg2) and $seg2 =~ $seg1){
            if($debug){print "found $seg1 in b1arr/tpl $seg2\n";}
            $found = 1; 
            last;
         }
        }
      }
    } 
    if(!$found){ 
       my $found2;
       foreach(@missing){
        if($_ eq $seg1) {$found2 = 1; last;}
       }
       if(!$found2){push(@missing, $seg1);}
    }
   }
    my ($out, $count, $var);
    $var = "Variable";
    foreach(@missing){
      $out .= $_.' ';
      if(++$count > 1){$var = "Variables";}
    }
    if(length($out)){ print_message( "Warning: $var $out missing from alphabet\n");}
    else {print_message( "All variables found in alphabet\n");}
}
#
# Utility for variable check
#
sub check_for_double_element($$$){
  my ($arr, $elem, $name) = @_;
  my @sa = split(/\s+/, $arr);
  foreach(@sa){
     if($_ eq $elem){
         print_message( "Element $elem twice in $name\n");
     }
  }
}
#
# print out a hash for debug purposes
#      
sub print_hash($$$){
  my ($hadr,$msg, $smsg) = @_;
  my %hash = %$hadr;
  my $size = keys(%hash);
  if($size){
  print "-----------------------$msg\n";
  my ($name, $elements);
  while (($name, $elements) = each(%hash))
  {
	print "$name has $elements\n";
  }
  } else{
    if($smsg){print "array $msg has size 0\n";}
  }
}
#
# is var in the expanded part of a #..# variable?
#
sub is_in_hash($$){
  my ($hadr, $var) = @_;
  my $debug = 0;
  my %hash = %$hadr;
  my $size = keys(%hash);
  my $found;
  my $seg = '<'.$var.'>';
  if($size){
    my ($name, $elements);
    while (($name, $elements) = each(%hash))
    {
       if($elements =~ /$seg/){
            $found = 1;
            last;
        }
# 	   if($debug){print "$name has $elements\n";}
    }
   }
   return $found;
}
  
sub check_hash_variables(){
  my $debug = 0;
  my $linenr;
  my (%avars, %hvars, %dvars, $avar, $hvar, $dvar, $avarname, $hvarname, $dvarname);
  foreach(@input){
    chomp;
    my @arr = split(//, $_);
    foreach(@arr){
       if($_ eq '<'){
           $avar = 1;
           next;
       }
       if($_ eq '>'){
           $avar = 0;
           if(defined($avars{$avarname})){
               ++$avars{$avarname};
           } else{
              $avars{$avarname} = 1;
           }
           $avarname = '';
           next;
       }
       if($_ eq '#' and !$hvar){
           $hvar = 1;
           next;
       }
       if($_ eq '$' and !$dvar){
           $dvar = 1;
           next;
       }
       if($_ eq '#' and $hvar){
           $hvar = 0;
           if(defined($hvars{$hvarname})){
               ++$hvars{$hvarname};
           } else{
              $hvars{$hvarname} = 1;
           }
           $hvarname = '';
           next;
       }
       if($_ eq '$' and $dvar){
           $dvar = 0;
           if(defined($dvars{$dvarname})){
               ++$dvars{$dvarname};
           } else{
              $dvars{$dvarname} = 1;
           }
           $dvarname = '';
           next;
       }
       if($avar){
         $avarname .= $_;
       } elsif($hvar){
         $hvarname .= $_;
       } elsif($dvar){
         $dvarname .= $_;
       }
  }
  ++$linenr;
 }
  my ($edvars,$ehvars, $eavars);
   ($edvars,$ehvars, $eavars) = read_evars($edvars,$ehvars, $eavars);
      
  my ($name, $elements);
 #   print "name:$name \nedv:$edvars \nehv:$ehvars\neav:$eavars\n";
  while (($name, $elements) = each(%dvars))
  {
    if($elements == 1 and $donotchkvar >= 2  and ($edvars eq '' or (good($name) and $edvars !~ /$name/ ))){
	   print_message("Warning: dvars $name has $elements element\n");
    }
  }
  while (($name, $elements) = each(%hvars))
  {
    if($elements == 1 and $donotchkvar >= 3  and ($ehvars eq '' or (good($name) and  $ehvars !~ /$name/))){
	   print_message("Warning: hvars $name has $elements element\n");
    }
  }
  while (($name, $elements) = each(%avars))
  {
    if($elements == 1 and $donotchkvar >= 4 and ($eavars eq '' or (good($name) and $eavars !~ /$name/ ))){
	   print_message("Warning: avars $name has $elements element\n");
    }
  }
}

